# Copyright 2022 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from termcolor import colored

from .LoggerInterface import LoggerInterface


class Logger(LoggerInterface):
    """ A console logger
        Prints the messages to the default console
    """

    def __init__(self, active: bool = True, name: str = "Logger", info_color: str = "magenta", show_init: bool = True):
        """ Initialize the logger

        Args:
            active (bool, optional): Defaults to True.
            name (str, optional): Defaults to "Logger".
            info_color (str, optional): Defaults to "magenta". The color of the info message
            show_init (bool, optional): Defaults to True. Show the initialization message
        """
        self.active = active
        self.name = name
        self.info_color = info_color

        if show_init:
            sep = "=" * 30
            time = datetime.now().strftime("%H:%M:%S")
            print(colored(f"[INFO] [{time}] [{self.name}] {sep}", self.info_color))
            print(colored(f"[INFO] [{time}] [{self.name}] {self.name} initialized", self.info_color))
            print(colored(f"[INFO] [{time}] [{self.name}] {sep}", self.info_color))

    def log(self, message: str):
        """ Log a message to the console
            Args:
                message (str): the message to log
        """
        if self.active:
            time = datetime.now().strftime("%H:%M:%S")
            print(colored(f"[INFO] [{time}] [{self.name}] {message}", self.info_color))

    def warn(self, message: str):
        if self.active:
            time = datetime.now().strftime("%H:%M:%S")
            print(colored(f"[WARNING] [{time}] [{self.name}] {message}", "yellow"))

    def error(self, message: str):
        if self.active:
            time = datetime.now().strftime("%H:%M:%S")
            print(colored(f"[ERROR] [{time}] [{self.name}] {message}", "red"))

    def debug(self, message: str):
        if self.active:
            time = datetime.now().strftime("%H:%M:%S")
            print(colored(f"[DEBUG] [{time}] [{self.name}] {message}", "blue"))

    def success(self, message: str):
        if self.active:
            time = datetime.now().strftime("%H:%M:%S")
            print(colored(f"[SUCCESS] [{time}] [{self.name}] {message}", "green"))



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import numpy as np
from enum import Enum

from .Autoencoder import Autoencoder
from .SSNP import SSNP
from ..AbstractDBM import AbstractDBM, DBM_DEFAULT_RESOLUTION, DEFAULT_TRAINING_EPOCHS, DEFAULT_BATCH_SIZE, FAST_DBM_STRATEGIES

from ...utils import track_time_wrapper, TEST_DATA_POINT_MARKER, TRAIN_DATA_POINT_MARKER
from ...Logger import LoggerInterface, Logger

time_tracker_console = Logger(name="Decision Boundary Mapper - DBM", info_color="cyan", show_init=False)

DEFAULT_MODEL_PATH = os.path.join("tmp", "SDBM")

class NNArchitecture(Enum):
    AUTOENCODER = "autoencoder"
    SSNP = "ssnp"


class SDBM(AbstractDBM):
    """
        Self Decision Boundary Mapper (SDBM)
        
        This class can generate a decision boundary map (DBM), without any need to provide a projection method.
        
        Public methods:
            fit: Learns a direct and an inverse projection by training a neural network. \n
            get_decision_boundary_map: Returns the decision boundary map for the given classifier. \n

        Example:
            >>> from SDBM import SDBM
            >>> import matplotlib.pyplot as plt
            >>> classifier = ...
            >>> sdbm = SDBM(classifier)
            >>> img, img_confidence, encoded_2d_train, encoded_2d_test = sdbm.get_decision_boundary_map(Xnd_train, Xnd_test, X2d_train, X2d_test)
            >>> plt.imshow(img)
            >>> plt.show()
    
    """

    def __init__(self, classifier, logger: LoggerInterface | None = None):
        """
        Initialize an Self Decision Boundary Mapper

        Args:
            classifier (tensorflow.keras.Model): The classifier to be used for the generation of SDBM.
            logger (LoggerInterface, optional): The logger class to be used for outputting the info messages. Defaults to None.
        """
        super().__init__(classifier, logger)
        self.neural_network = None  # type: ignore

    @track_time_wrapper(logger=time_tracker_console)
    def fit(self,
            X: np.ndarray, Y: np.ndarray,
            architecture: NNArchitecture = NNArchitecture.AUTOENCODER,
            epochs: int = DEFAULT_TRAINING_EPOCHS, batch_size: int = DEFAULT_BATCH_SIZE,
            load_folder: str = DEFAULT_MODEL_PATH, is_data_normalized: bool = True):
        """
        Train a neural network that will contain the direct projection and the inverse projection.
        This neural network will be used to reduce the dimensionality of the data (nD -> 2D) and decode the 2D space to nD.

        Args:
            X (np.ndarray): Training data
            Y (np.ndarray): Training labels
            architecture (NNArchitecture): The architecture of the neural network. Defaults to NNArchitecture.AUTOENCODER.
            epochs (int, optional): The number of epochs to train the neural network. Defaults to 300.
            batch_size (int, optional): Defaults to 32.
            load_folder (str, optional): The folder path which contains a pre-trained network or will be used to store it if not exists. Defaults to DEFAULT_MODEL_PATH.
            is_data_normalized (bool, optional): Determine the last layer activation function of the decoder, sigmoid or relu. Defaults to True (i.e. activation sigmoid).
     
        Returns:
            neural_network (Autoencoder | SSNP): The trained neural network.
        """
        match architecture:
            case NNArchitecture.SSNP:
                ssnp = SSNP(folder_path=load_folder, logger=self.console)
                ssnp.fit(X, Y, epochs, batch_size, is_data_normalized=is_data_normalized)
                return ssnp
            case _:
                autoencoder = Autoencoder(folder_path=load_folder, logger=self.console)
                autoencoder.fit(X, epochs, batch_size, is_data_normalized=is_data_normalized)
                return autoencoder

    def generate_boundary_map(self,
                              X_train: np.ndarray, Y_train: np.ndarray,
                              X_test: np.ndarray, Y_test: np.ndarray,
                              nn_train_epochs: int = DEFAULT_TRAINING_EPOCHS, 
                              nn_train_batch_size: int = DEFAULT_BATCH_SIZE,
                              nn_architecture: NNArchitecture = NNArchitecture.AUTOENCODER,
                              resolution: int = DBM_DEFAULT_RESOLUTION,
                              fast_decoding_strategy: FAST_DBM_STRATEGIES = FAST_DBM_STRATEGIES.NONE,
                              load_folder: str = DEFAULT_MODEL_PATH,
                              is_data_normalized: bool = True,
                              ):
        """Generate the decision boundary map

            Args:
                X_train (np.ndarray): Training data
                Y_train (np.ndarray): Training labels
                X_test (np.ndarray): Testing data
                Y_test (np.ndarray): Testing labels
                nn_train_epochs (int, optional): The number of epochs to train the neural network that will give the direct and inverse projections. Defaults to 300.
                nn_train_batch_size (int, optional): Defaults to 32.
                nn_architecture (NNArchitecture): The architecture to use for the SDBM. Defaults to NNArchitecture.AUTOENCODER
                resolution (int, optional): The resolution of the decision boundary map. Defaults to DBM_DEFAULT_RESOLUTION = 256.
                fast_decoding_strategy (FAST_DBM_STRATEGIES, optional): The strategy to use for the generation of the DBM. Defaults to FAST_DBM_STRATEGIES.NONE
                load_folder (str, optional): The folder path which contains a pre-trained neural network or in which it will be stored. Defaults to DEFAULT_MODEL_PATH.
                is_data_normalized (bool, optional): Determine the last layer activation function of the decoder, sigmoid or relu. Defaults to True (i.e. activation sigmoid).
     
            Returns:
                img (np.ndarray): The decision boundary map
                img_confidence (np.ndarray): The confidence map
                encoded_2d_train (np.ndarray): The 2D coordinates of the training data for each pixel of the decision boundary map and the assigned classifier label
                encoded_2d_test (np.ndarray): The 2D coordinates of the testing data for each pixel of the decision boundary map and the assigned classifier label
                
            Example:
                >>> import SDBM
                >>> classifier = build_classifier(...)
                >>> sdbm = SDBM.SDBM(classifier)
                >>> img, img_confidence, _, _ = sdbm.generate_boundary_map(X_train, Y_train, X_test, Y_test)
                >>> plt.imshow(img)
                >>> plt.show()
        """

        # first train the autoencoder if it is not already trained
        if self.neural_network is None:
            X = np.concatenate((X_train, X_test), axis=0)
            Y = np.concatenate((Y_train, Y_test), axis=0)
            self.neural_network = self.fit(X, Y,
                                           architecture=nn_architecture,
                                           load_folder=load_folder,
                                           epochs=nn_train_epochs,
                                           batch_size=nn_train_batch_size,
                                           is_data_normalized=is_data_normalized
                                           )

        # encoder the train and test data and show the encoded data in 2D space
        self.console.log("Encoding the training data to 2D space")
        encoded_training_data = self.neural_network.encode(X_train)
        self.console.log("Encoding the testing data to 2D space")
        encoded_testing_data = self.neural_network.encode(X_test)
        # generate the 2D image in the encoded space
        self.console.log("Decoding the 2D space... 2D -> nD")

        self.resolution = resolution

        img, img_confidence = self.get_dbm(fast_decoding_strategy, resolution, load_folder)

        self.X2d = np.concatenate((encoded_training_data, encoded_testing_data), axis=0)
        self.Xnd = np.concatenate((X_train.reshape((X_train.shape[0], -1)), X_test.reshape((X_test.shape[0], -1))), axis=0)

        # transform the encoded data to be in the range [0, resolution)
        encoded_testing_data *= (resolution - 1)
        encoded_training_data *= (resolution - 1)
        encoded_training_data = encoded_training_data.astype(int)
        encoded_testing_data = encoded_testing_data.astype(int)

        encoded_2d_train = np.zeros((len(encoded_training_data), 3))
        encoded_2d_test = np.zeros((len(encoded_testing_data), 3))

        for k in range(len(encoded_training_data)):
            [i, j] = encoded_training_data[k]
            encoded_2d_train[k] = [i, j, img[i, j]]
        for k in range(len(encoded_testing_data)):
            [i, j] = encoded_testing_data[k]
            encoded_2d_test[k] = [i, j, img[i, j]]

        for [i, j] in encoded_testing_data:
            img[i, j] = TEST_DATA_POINT_MARKER
            img_confidence[i, j] = 1
        for [i, j] in encoded_training_data:
            img[i, j] = TRAIN_DATA_POINT_MARKER
            img_confidence[i, j] = 1

        return (img, img_confidence, encoded_2d_train, encoded_2d_test)

    def _predict2dspace_(self, X2d: np.ndarray):
        """ 
        Predicts the labels for the given 2D data set.

        Args:
            X2d (np.ndarray): The 2D data set

        Returns:
            predicted_labels (np.array): The predicted labels for the given 2D data set
            predicted_confidence (np.array): The predicted probabilities for the given 2D data set
        """
        spaceNd = self.neural_network.decode(X2d, verbose=0)
        predictions = self.classifier.predict(spaceNd, verbose=0)                 # type: ignore
        predicted_labels = np.array([np.argmax(p) for p in predictions])
        predicted_confidence = np.array([np.max(p) for p in predictions])
        return predicted_labels, predicted_confidence, predictions



# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
from shutil import rmtree
import tensorflow as tf
import numpy as np

from .DBMPlotterGUI import DBMPlotterGUI
from ..DBM import SDBM, DBM, NNArchitecture
from ..Logger import Logger
from ..utils import import_dataset, import_mnist_dataset, import_cifar10_dataset, import_fashion_mnist_dataset, import_folder_dataset, generate_class_name_mapper, TRAIN_2D_FILE_NAME, TEST_2D_FILE_NAME

DBM_NNINV_TECHNIQUE = "nnInv"
SDBM_SSNP_TECHNIQUE = "ssnp"
SDBM_AUTOENCODER_TECHNIQUE = "autoencoder"

DBM_TECHNIQUES = {
    SDBM_AUTOENCODER_TECHNIQUE: SDBM,
    SDBM_SSNP_TECHNIQUE: SDBM,
    DBM_NNINV_TECHNIQUE: DBM,
}

CUSTOM_PROJECTION_TECHNIQUE = "CUSTOM"
PROJECTION_TECHNIQUES = [
    "t-SNE",
    "UMAP",
    "PCA",
    CUSTOM_PROJECTION_TECHNIQUE,
]

DATASETS_IMPORTERS = {
    "MNIST": import_mnist_dataset,
    "FASHION MNIST": import_fashion_mnist_dataset,
    "CIFAR10": import_cifar10_dataset,
}

DBM_FOLDER_NAME = "DBM"
SDBM_FOLDER_NAME = "SDBM"
HISTORY_FILE_NAME = "history.json"

TMP_FOLDER = os.path.join(os.getcwd(), "tmp")
SAMPLES_LIMIT = 5000  # Limit the number of samples to be loaded from the dataset
DEFAULT_DBM_RESOLUTION = 256

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'  # Disable tensorflow logs
"""
0 = all messages are logged (default behavior)
1 = INFO messages are not printed
2 = INFO and WARNING messages are not printed
3 = INFO, WARNING, and ERROR messages are not printed
"""


class GUIController:
    def __init__(self, window, gui, gui_logger):
        self.window = window
        self.gui = gui
        self.create_tmp_folder()
        self.logger = Logger(name="GUIController")
        self.gui_logger = gui_logger

        
        # --------------- Data set ----------------
        self.dataset_name = "Dataset"
        self.X_train, self.Y_train = None, None
        self.X_test, self.Y_test = None, None
        self.X_train_2d, self.X_test_2d = None, None

        # --------------- Classifier --------------
        self.classifier = None
        
        # --------------- Others ---------------------
        self.data_shape = (28, 28) # this is the shape of the data, it is used to reshape the data when importing it from a csv file
        self.class_names_mapper = lambda x: str(x)
        self.helper_encoder = None
        self.helper_decoder = None
       
        self.X_train_features = None 
        self.X_test_features = None

    def stop(self):
        self.logger.log("Clearing resources...")
        # self.logger.log("Removing tmp folder...")
        # self.remove_tmp_folder()
        
    def create_tmp_folder(self):
        if not os.path.exists(TMP_FOLDER):
            os.makedirs(TMP_FOLDER)

    def remove_tmp_folder(self):
        if os.path.exists(TMP_FOLDER):
            rmtree(TMP_FOLDER)

    def switch_visibility(self, elements, visible):
        for x in elements:
            self.window[x].update(visible=visible)
        self.window.refresh()

    def handle_select_neural_network_folder(self, folder):
        try:
            # Get list of files in folder
            file_list = os.listdir(folder)
        except:
            file_list = []

        fnames = [f for f in file_list
                  if (os.path.isfile(os.path.join(folder, f)) and (f.lower().endswith((".h5")))) or os.path.isdir(os.path.join(folder, f))]
        return fnames

    def handle_select_data_folder(self, folder):
        try:
            # Get list of files in folder
            file_list = os.listdir(folder)
        except:
            file_list = []

        ALLOWED_DATA_EXTENSIONS = [".csv", ".txt", ".npy"]
        fnames = [f for f in file_list
                  if os.path.isfile(os.path.join(folder, f)) and (os.path.splitext(f)[-1].lower() in ALLOWED_DATA_EXTENSIONS)]
        return fnames

    def handle_file_list_event(self, event, values):
        try:
            filename = os.path.join(values["-DATA FOLDER-"], values["-DATA FILE LIST-"][0])
            self.switch_visibility(["-DATA FILE TOUT PIN-"], True)
            self.window["-DATA FILE TOUT-"].update(filename)
        except Exception as e:
            self.logger.error("Error while loading data file" + str(e))

    def handle_classifier_file_list_event(self, event, values):
        try:
            filename = os.path.join(values["-CLASSIFIER FOLDER-"], values["-CLASSIFIER FILE LIST-"][0])
            self.window["-CLASSIFIER PATH TOUT-"].update(filename)
            self.switch_visibility(["-CLASSIFIER PATH TOUT PIN-"], True)
        except Exception as e:
            self.logger.error("Error while loading data file" + str(e))
            
    def handle_feature_extractor_file_list_event(self, event, values):
        try:
            filename = os.path.join(values["-FEATURE EXTRACTOR FOLDER-"], values["-FEATURE EXTRACTOR FILE LIST-"][0])
            self.window["-FEATURE EXTRACTOR PATH TOUT-"].update(filename)
            self.switch_visibility(["-FEATURE EXTRACTOR PATH TOUT PIN-"], True)
        except Exception as e:
            self.logger.error("Error while loading data file" + str(e))
        
    def handle_upload_classifier_event(self, event, values):
        try:
            fname = os.path.join(values["-CLASSIFIER FOLDER-"], values["-CLASSIFIER FILE LIST-"][0])
            self.classifier = tf.keras.models.load_model(fname)
            self.window["-CLASSIFIER PATH TOUT-"].update(fname)
            self.switch_visibility(["-CLASSIFIER PATH TOUT PIN-"], True)
            self.logger.log("Classifier loaded successfully")
            self.gui_logger.log("Classifier loaded successfully")
            if self.X_train is not None and self.Y_train is not None and self.X_test is not None and self.Y_test is not None:
                self.switch_visibility(["-DBM BTN-"], True)
        except Exception as e:
            self.logger.error("Error while loading classifier" + str(e))
    
    def handle_upload_feature_extractor_event(self, event, values):
        try:
            fname = os.path.join(values["-FEATURE EXTRACTOR FOLDER-"], values["-FEATURE EXTRACTOR FILE LIST-"][0])
            model = tf.keras.models.load_model(fname)
            if event == "-UPLOAD FEATURE EXTRACTOR ENCODER-":
                self.helper_encoder = model
                model_type = "encoder"
            elif event == "-UPLOAD FEATURE EXTRACTOR DECODER-":
                self.helper_decoder = model
                model_type = "decoder"
            else:
                raise Exception("DEVELOPMENT ISSUE, the event can only be -UPLOAD FEATURE EXTRACTOR ENCODER- or -UPLOAD FEATURE EXTRACTOR DECODER-")
            self.window["-FEATURE EXTRACTOR PATH TOUT-"].update(fname)
            self.logger.log(f"Feature extractor {model_type} loaded successfully")
            self.gui_logger.log(f"Feature extractor {model_type} loaded successfully")
        except Exception as e:
            self.logger.error("Error while loading feature extractor " + fe_type + " " + str(e))
    
    def handle_upload_train_data_event(self, event, values):
        try:
            filename = os.path.join(values["-DATA FOLDER-"], values["-DATA FILE LIST-"][0])
            self.X_train, self.Y_train = import_dataset(filename, limit=int(0.7*SAMPLES_LIMIT), shape=self.data_shape)
           
            self.X_train = self.X_train.astype("float32") / 255
            self.Y_test = self.Y_train.astype("int")
           
            self.num_classes = np.unique(self.Y_train).shape[0]
            if self.classifier is not None and self.X_test is not None and self.Y_test is not None:
                self.switch_visibility(["-DBM BTN-"], True)

            self.window["-TRAIN DATA FILE-"].update("Training data file: " + filename)
            self.window["-TRAIN DATA SHAPE-"].update(f"Training data shape: X {self.X_train.shape} Y {self.Y_train.shape}")
        except Exception as e:
            self.gui_logger.error("Error while loading data file" + str(e))

    def handle_upload_test_data_event(self, event, values):
        try:
            filename = os.path.join(values["-DATA FOLDER-"], values["-DATA FILE LIST-"][0])
            self.X_test, self.Y_test = import_dataset(filename, limit=int(0.3*SAMPLES_LIMIT), shape=self.data_shape)
            self.X_test = self.X_test.astype("float32") / 255
            self.Y_test = self.Y_test.astype("int")
            
            if self.classifier is not None and self.X_train is not None and self.Y_train is not None:
                self.switch_visibility(["-DBM BTN-"], True)

            self.window["-TEST DATA FILE-"].update("Testing data file: " + filename)
            self.window["-TEST DATA SHAPE-"].update(f"Testing data shape: X {self.X_test.shape} Y {self.Y_test.shape}")
        except Exception as e:
            self.gui_logger.error("Error while loading data file" + str(e))

    def handle_upload_2d_data_event(self, event, values):
        try:
            filename = os.path.join(values["-DATA 2D FOLDER-"], values["-DATA 2D FILE LIST-"][0])
            if event == "-UPLOAD 2D TRAIN DATA BTN-":
                self.X_train_2d, _ = import_dataset(filename, labels_index=None)
                self.gui_logger.log("2D train data loaded successfully")
            elif event == "-UPLOAD 2D TEST DATA BTN-":
                self.X_test_2d, _ = import_dataset(filename, labels_index=None)
                self.gui_logger.log("2D test data loaded successfully")
                
        except Exception as e:
            self.gui_logger.error("Error while loading data file" + str(e))

    def handle_upload_known_data_event(self, event, values):
        self.dataset_name = event[len("-UPLOAD "):-len(" DATA BTN-")]
        
        (X_train, Y_train), (X_test, Y_test) = DATASETS_IMPORTERS[self.dataset_name]()

        X_train = X_train.astype('float32') / 255
        X_test = X_test.astype('float32') / 255

        X_train, Y_train = X_train[:int(0.7*SAMPLES_LIMIT)], Y_train[:int(0.7*SAMPLES_LIMIT)]
        X_test, Y_test = X_test[:int(0.3*SAMPLES_LIMIT)], Y_test[:int(0.3*SAMPLES_LIMIT)]

        self.X_train, self.Y_train, self.X_test, self.Y_test = X_train, Y_train, X_test, Y_test
        self._post_uploading_processing_()
    
    def handle_upload_folder_dataset_event(self, event, values):
        folder = values["-DATA FOLDER-"]
        self.dataset_name = "_".join(folder.split(os.sep)[-2:])
        (X_train, Y_train), (X_test, Y_test) = import_folder_dataset(folder)

        X_train = X_train.astype('float32') / 255
        X_test = X_test.astype('float32') / 255

        self.X_train, self.Y_train, self.X_test, self.Y_test = X_train, Y_train, X_test, Y_test
        
        # Not supporting class names mapper because the dataset can be structured in a lot of different formats
        #self.class_names_mapper = generate_class_name_mapper(os.path.join(folder, "..", "classes.txt"))
        
        self._post_uploading_processing_()
   
    def _post_uploading_processing_(self):
        self.num_classes = np.unique(self.Y_train).shape[0]

        self.window["-TRAIN DATA FILE-"].update(f"Training data: {self.dataset_name}")
        self.window["-TRAIN DATA SHAPE-"].update(f"Training data shape: X {self.X_train.shape} Y {self.Y_train.shape}")

        self.window["-TEST DATA FILE-"].update(f"Testing data: {self.dataset_name}")
        self.window["-TEST DATA SHAPE-"].update(f"Testing data shape: X {self.X_test.shape} Y {self.Y_test.shape}")
        
        self.window["-UPLOAD DATA COLLAPSABLE-"].update(visible=False)
        
        if self.classifier is not None:
            self.switch_visibility(["-DBM BTN-"], True)

    def fetch_2d_data_from_folder(self, folder_path):
        if not os.path.exists(os.path.join(folder_path, TRAIN_2D_FILE_NAME)):
            X_train_2d = None
        else:
            with open(os.path.join(folder_path, TRAIN_2D_FILE_NAME), "rb") as f:
                X_train_2d = np.load(f)
        if not os.path.exists(os.path.join(folder_path, TEST_2D_FILE_NAME)):
            X_test_2d = None
        else:
            with open(os.path.join(folder_path, TEST_2D_FILE_NAME), "rb") as f:
                X_test_2d = np.load(f)
        return X_train_2d, X_test_2d

    def handle_get_decision_boundary_mapping_event(self, event, values):
        if self.classifier is None:
            self.gui_logger.error("No classifier provided, impossible to generate the DBM...")
            return

        if self.X_train is None or self.Y_train is None or self.X_test is None or self.Y_test is None:
            self.gui_logger.error("Data is incomplete impossible to generate the DBM...")
            return

        if self.helper_decoder is not None and self.helper_encoder is not None and self.X_train_features is None and self.X_test_features is None:
            self.gui_logger.log("Extracting the features from the data using the provided feature encoder, please wait...")
            self.X_train_features = self.helper_encoder.predict(self.X_train, verbose=0)
            self.X_test_features = self.helper_encoder.predict(self.X_test, verbose=0)
            self.gui_logger.log("Features extracted successfully using the feature encoder.")
       
        X_train = self.X_train if self.X_train_features is None else self.X_train_features
        X_test = self.X_test if self.X_test_features is None else self.X_test_features

        # update loading state
        self.switch_visibility(["-DBM IMAGE-"], False)
        self.switch_visibility(["-DBM TEXT-", "-DBM IMAGE LOADING-"], True)

        dbm = DBM_TECHNIQUES[values["-DBM TECHNIQUE-"]](classifier=self.classifier, logger=self.gui_logger)

        projection_technique = values["-PROJECTION TECHNIQUE-"]
        resolution = DEFAULT_DBM_RESOLUTION  # values["-DBM IMAGE RESOLUTION INPUT-"]

        self.gui_logger.log(f"DBM resolution: {resolution}")

        save_folder = os.path.join("tmp", self.dataset_name)
        dbm_technique = values["-DBM TECHNIQUE-"]
        
        if dbm_technique == DBM_NNINV_TECHNIQUE:
            save_folder = os.path.join(save_folder, DBM_FOLDER_NAME)
            load_folder = os.path.join(save_folder, projection_technique)
            
            if projection_technique == CUSTOM_PROJECTION_TECHNIQUE:
                if self.X_train_2d is None or self.X_test_2d is None:
                    raise Exception(f"No 2D data provided, impossible to generate the DBM for projection technique {CUSTOM_PROJECTION_TECHNIQUE}...")
            else:
                self.X_train_2d, self.X_test_2d = self.fetch_2d_data_from_folder(load_folder)

            dbm_info = dbm.generate_boundary_map(
                Xnd_train=X_train,
                Xnd_test=X_test,
                X2d_train=self.X_train_2d,
                X2d_test=self.X_test_2d,
                resolution=resolution,
                load_folder=save_folder,
                projection=projection_technique
            )
        else:
            save_folder = os.path.join(save_folder, SDBM_FOLDER_NAME)
            projection_technique = None
            dbm_info = dbm.generate_boundary_map(
                X_train=X_train,
                Y_train=self.Y_train,
                X_test=X_test,
                Y_test=self.Y_test,
                nn_architecture=NNArchitecture(dbm_technique),
                resolution=resolution,
                load_folder=save_folder,
            )

        img, img_confidence, encoded_training_data, encoded_testing_data = dbm_info
        
        show_dbm_history = values["-DBM HISTORY CHECKBOX-"]
        if show_dbm_history:
            path = os.path.join(save_folder, HISTORY_FILE_NAME)
            training_history = None
            if os.path.exists(path):   
                with open(path, 'r') as f:
                    training_history = json.load(f)
            self.gui.show_dbm_history(training_history)

        # ---------------------------------
        # create a GUI window for Decision Boundary Mapping
        return DBMPlotterGUI(
            dbm_model=dbm,
            img=img,
            img_confidence=img_confidence,
            encoded_train=encoded_training_data,
            encoded_test=encoded_testing_data,
            X_train=self.X_train,
            Y_train=self.Y_train,
            X_test=self.X_test,
            Y_test=self.Y_test,
            X_train_2d=self.X_train_2d,
            X_test_2d=self.X_test_2d,
            main_gui=self.gui,  # reference to the main GUI
            save_folder=save_folder,
            projection_technique=projection_technique,
            class_name_mapper=self.class_names_mapper,
            helper_decoder=self.helper_decoder,
            X_train_latent=self.X_train_features,
            X_test_latent=self.X_test_features
        )
        


# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from termcolor import colored

from .LoggerInterface import LoggerInterface


class LoggerGUI(LoggerInterface):
    """ Logs the messages to the GUI.
    """

    def __init__(self, name: str = "Logger", active: bool = True, output=None, update_callback=lambda: "", info_color: str = "magenta", show_init: bool = True):
        """ Initialize the logger

        Args:
            name (str, optional): Defaults to "Logger".
            active (bool, optional): Defaults to True.
            output (any, optional): The GUI reference where the logger should display messages. Defaults to None.
            update_callback (python function, optional): The function that is called to update the GUI. Defaults to lambda x:x.
        """
        super().__init__()

        self.active = active
        self.name = name
        self.info_color = info_color

        if output is not None:
            self.output = output
            self.update_callback = update_callback
        else:
            print("No output provided for LoggerGUI. LoggerGUI will not print anything.")
            self.active = False

        if show_init and self.active:
            sep = "=" * 30
            time = datetime.now().strftime("%H:%M:%S")

            self.print(f"[INFO] [{time}] {sep}", self.info_color)
            self.print(f"[INFO] [{time}] {self.name} initialized", self.info_color)
            self.print(f"[INFO] [{time}] {sep}", self.info_color)

    def print(self, message: str, color: str = 'magenta'):
        if self.active:
            self.output.print(message, text_color=color)
            self.update_callback()

    def log(self, message: str):
        time = datetime.now().strftime("%H:%M:%S")
        self.print(f"[INFO] [{time}] {message}", self.info_color)

    def warn(self, message: str):
        time = datetime.now().strftime("%H:%M:%S")
        self.print(f"[WARNING] [{time}] {message}", "orange")

    def error(self, message: str):
        time = datetime.now().strftime("%H:%M:%S")
        self.print(f"[ERROR] [{time}] {message}", "red")

    def debug(self, message: str):
        time = datetime.now().strftime("%H:%M:%S")
        self.print(f"[DEBUG] [{time}] {message}", "blue")

    def success(self, message: str):
        time = datetime.now().strftime("%H:%M:%S")
        self.print(f"[SUCCESS] [{time}] {message}", "green")



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import numpy as np
import os

from ..AbstractNN import AbstractNN, SEED
from ...Logger import LoggerInterface, LoggerModel

DEFAULT_MODEL_PATH = os.path.join("tmp", "DBM")
NNINV_NAME = "NNInv"
DECODER_NAME = "decoder"
DECODER_LOSS = "mean_squared_error"


class NNInv(AbstractNN):
    """
        Inverse Projection Neural Network.        
        The inverse projection 2D -> nD. 
        (Sequential model 2 -> 32 -> 64 -> 128 -> 512 -> nD)            
    """

    def __init__(self,
                 logger: LoggerInterface | None = None,
                 folder_path: str = DEFAULT_MODEL_PATH):
        """
            Creates an inverse Projection Neural Network model.
        """
        super().__init__(folder_path=folder_path, nn_name=NNINV_NAME, logger=logger)

    def __build__(self, output_shape: tuple = (2, 2), show_summary: bool = False, is_data_normalized: bool = True):
        """
        Builds an NNInv model
        (Sequential 2D -> 32 -> 64 -> 128 -> 512 -> nD)

        Args:
            output_shape (tuple, optional): The output shape of the Nd data. Defaults to (2,2).
            show_summary (bool, optional): If True, the model summary will be printed. Defaults to False.
            is_data_normalized (bool, optional): Determine the last layer activation function, sigmoid or relu. Defaults to True (i.e. activation sigmoid).
        """

        # computing the output size
        output_size = 1
        for i in range(len(output_shape)):
            output_size *= output_shape[i]

        last_layer_activation_function = 'sigmoid' if is_data_normalized else 'relu'

        self.decoder = tf.keras.Sequential([
            tf.keras.layers.Dense(32, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  kernel_regularizer=tf.keras.regularizers.l2(0.0002)),
            tf.keras.layers.Dense(64, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(128, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(512, activation='relu', kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED),  # type: ignore
                                  bias_initializer=tf.keras.initializers.Constant(0.01)),  # type: ignore
            tf.keras.layers.Dense(output_size, activation=last_layer_activation_function,
                                  kernel_initializer=tf.keras.initializers.HeUniform(seed=SEED)),  # type: ignore
            tf.keras.layers.Reshape(output_shape)
        ], name=DECODER_NAME)

        input_layer = tf.keras.Input(shape=(2,), name="input")

        self.neural_network = tf.keras.models.Model(inputs=input_layer,
                                                    outputs=[self.decoder(input_layer)],
                                                    name=NNINV_NAME)

        self.neural_network.compile(optimizer=tf.keras.optimizers.Adam(),
                                    loss=DECODER_LOSS,
                                    metrics=["accuracy"])

        if show_summary:
            self.neural_network.summary(print_fn=self.console.log)

    def fit(self,
            x2d: np.ndarray, xNd: np.ndarray,
            epochs: int = 300, batch_size: int = 32,
            is_data_normalized: bool = True,
            ):
        """ 
        Fits the model to the specified data.

        Args:
            x2d (np.ndarray): Train input values (2D)
            xNd (np.ndarray): Train input values (nD)
            epochs (int, optional): The number of epochs. Defaults to 300.
            batch_size (int, optional): Data points used for one batch. Defaults to 32.
            is_data_normalized (bool, optional): Determine the last layer activation function, sigmoid or relu. Defaults to True (i.e. activation sigmoid).
        """
        if self.neural_network is not None:
            self.console.log("Model already loaded. Skipping build.")
            return

        self.console.log("Building model according to the data shape.")
        self.__build__(output_shape=xNd.shape[1:], show_summary=True, is_data_normalized=is_data_normalized)

        stopping_callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', mode='min', patience=20, restore_best_weights=True)
        logger_callback = LoggerModel(name=NNINV_NAME, show_init=False, epochs=epochs, print_fn=self.console.log)
        self.console.log("Fitting model...")

        hist = self.neural_network.fit(x2d, xNd,
                                       epochs=epochs,
                                       batch_size=batch_size,
                                       shuffle=True,
                                       validation_split=0.2,
                                       callbacks=[stopping_callback,
                                                  logger_callback],
                                       verbose=0)

        self.console.log("Model fitted!")
        self.save(hist)
        # self.show_predictions(dataNd=xNd_test, data2d=x2d_test, labels=y_test)



# Copyright 2023 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from umap import UMAP
import numpy as np


def generate_transformation(projection: TSNE | PCA | UMAP):
    """
    Generates a 2D projection functionality
    
    Returns:
        function: X: (nd data) -> 2d projection
    """
    def transform(X: np.ndarray):
        """ 
        Transforms the given data to 2D using the projection method.

        Args:
            X (np.ndarray): The data to be transformed
        """
        X_flat = X.reshape(X.shape[0], -1)
        X2d = projection.fit_transform(X_flat)
        return X2d
    
    return transform

PROJECTION_METHODS = {
    "t-SNE": generate_transformation(TSNE(n_components=2, random_state=0, learning_rate="auto", init="random")),
    "PCA": generate_transformation(PCA(n_components=2, random_state=0)),
    "UMAP": generate_transformation(UMAP(n_components=2, random_state=0)),
    "CUSTOM": None,
}



# Copyright 2022 Cristian Grosu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from keras.datasets import fashion_mnist, mnist, cifar10
import os
import pandas as pd
from PIL import Image

from .. import Logger

ALLOWED_DATA_FORMATS = [".csv", ".txt", ".npy"]


def import_mnist_dataset() -> tuple:
    """Imports the MNIST dataset from keras.datasets.mnist

    Returns:
        (X_train, Y_train), (X_test, Y_test): The train and test sets
    """
    console = Logger(name="MNIST dataset importer")
    (train_X, train_y), (test_X, test_y) = mnist.load_data()
    console.log("MNIST dataset imported")
    console.log(f"Train set: {train_X.shape}")
    console.log(f"Test set: {test_X.shape}")
    return (train_X, train_y), (test_X, test_y)


def import_fashion_mnist_dataset() -> tuple:
    """Imports the FASHION MNIST dataset from keras.datasets.fashion_mnist

    Returns:
        (X_train, Y_train), (X_test, Y_test): The train and test sets
    """
    console = Logger(name="FASHION MNIST dataset importer")
    (train_X, train_y), (test_X, test_y) = fashion_mnist.load_data()
    console.log("FASHION MNIST dataset imported")
    console.log(f"Train set: {train_X.shape}")
    console.log(f"Test set: {test_X.shape}")
    return (train_X, train_y), (test_X, test_y)


def import_cifar10_dataset() -> tuple:
    """Imports the CIFAR10 dataset from keras.datasets.cifar10

    Returns:
        (X_train, Y_train), (X_test, Y_test): The train and test sets
    """
    console = Logger(name="CIFAR10 dataset importer")
    (train_X, train_y), (test_X, test_y) = cifar10.load_data()
    console.log("CIFAR10 dataset imported")
    console.log(f"Train set: {train_X.shape}")
    console.log(f"Test set: {test_X.shape}")
    return (train_X, train_y), (test_X, test_y)

def import_folder_dataset(dir:str, train_test_split = 0.8) -> tuple:
    console = Logger(name="Dataset importer")
    
    # first check if the dataset is already splitted into train and test sets
    if "train" in os.listdir(dir) and "test" in os.listdir(dir):
        # import the training set
        train_dir = os.path.join(dir, "train")
        file_list = [file for file in os.listdir(train_dir) if file.endswith(".png") or file.endswith(".jpg")]
        file_list.sort()
        X_train = np.array([np.array(Image.open(os.path.join(train_dir, fname))) for fname in file_list])
        # if there is a labels file provided, use that file, otherwise use the names of the files
        if "y.npy" in os.listdir(train_dir):
            Y_train = np.load(os.path.join(train_dir, "y.npy"))
        else:
            Y_train = np.array([int(fname.split(".")[0].split("_")[0]) for fname in file_list]) 
        # import the test set 
        test_dir = os.path.join(dir, "test")
        file_list = [file for file in os.listdir(test_dir) if file.endswith(".png") or file.endswith(".jpg")]
        file_list.sort()
        X_test = np.array([np.array(Image.open(os.path.join(test_dir, fname))) for fname in file_list])
        # if there is a labels file provided, use that file, otherwise use the names of the files
        if "y.npy" in os.listdir(test_dir):
            Y_test = np.load(os.path.join(test_dir, "y.npy"))
        else:
            Y_test = np.array([int(fname.split(".")[0].split("_")[0]) for fname in file_list]) 
        # post processing the labels
        y_train_min, y_test_min = np.min(Y_train), np.min(Y_test)
        if y_train_min != 0:
            Y_train -= y_train_min
        if y_test_min != 0:
            Y_test -= y_test_min
        # ----------------------------------------------------------------
        console.log(f"Dataset imported from {dir}")
        console.log(f"Train set: {X_train.shape}")
        console.log(f"Test set: {X_test.shape}")
        
        return (X_train, Y_train), (X_test, Y_test)

    # if the dataset is in one folder format then import the whole dataset and use the train_test_split
    file_list = [os.path.join(dir, file) for file in os.listdir(dir)]
    Y = np.array([int(fname.split("_")[0]) - 1 for fname in os.listdir(dir)]) 
    X = np.array([np.array(Image.open(fname)) for fname in file_list])
    console.log(f"Dataset imported from {dir}")
    console.log(f"Dataset shape: X {X.shape}, Y {Y.shape}")
    
    size = len(Y)

    (X_train, X_test) = X[:int(train_test_split*size)], X[int(train_test_split*size):]
    (Y_train, Y_test) = Y[:int(train_test_split*size)], Y[int(train_test_split*size):]
    
    console.log(f"Train set: {X_train.shape}")
    console.log(f"Test set: {X_test.shape}")
    return (X_train, Y_train), (X_test, Y_test)

    
def import_dataset(file_path:str,
                   labels_index: int | None = 0,
                   limit: int | None = None,
                   shape: tuple | None = None,
                   ) -> tuple:       
    """Imports a dataset from a file"""
    if not os.path.exists(file_path):
        print("File not found")
        return None, None
    
    if not os.path.splitext(file_path)[-1] in ALLOWED_DATA_FORMATS:
        print("File format not supported, please use one of the following: ", ALLOWED_DATA_FORMATS)
        return None, None
     
    data = None   
    if os.path.splitext(file_path)[-1] == ".csv" or os.path.splitext(file_path)[-1] == ".txt":
        data = import_csv_dataset(file_path, limit=limit)

    if os.path.splitext(file_path)[-1] == ".npy":
        data = import_npy_dataset(file_path, limit=limit)
    
    if data is None:
        return None, None
    
    if labels_index is None:
        return data, None
 
    
    Y = data[:, labels_index]
    X = np.delete(data, labels_index, axis=1)
      
    if shape is not None:
        X = X.reshape(X.shape[0], *shape)
    return X, Y
  
def import_csv_dataset(file_path: str,
                       limit: int | None = None) -> np.ndarray:
    """Imports a dataset from a csv file

    Args:
        file_path (str): The file path
        limit (_type_, optional): The limit of data points to be loaded. Defaults to None.
    Returns:
        X (np.ndarray): The data points and labels
    """
    data_frame = pd.read_csv(file_path, nrows=limit, skiprows = 1, header=None)
    return np.array(data_frame.values)
   

def import_npy_dataset(file_path: str,
                        limit: int | None = None) -> np.ndarray:
    """Imports a dataset from a npy file"""
    with open(file_path, 'rb') as f:
        data = np.load(f)
    X = data[:limit]
    return X
